@@ -135,9 +135,9 @@ static int _mmc_send_cmd(MMC *mmc, const uint8_t *cmd, uint8_t *buf, size_t tx,
         DEBUG(DBG_MMC, "Send LINUX MMC cmd %s: (%p)\n",
               print_hex(str, cmd, 16), mmc);
         if (tx) {
-            DEBUG(DBG_MMC, "  Buffer: %s -> (%p)\n", print_hex(str, buf, tx), mmc);
+            DEBUG(DBG_MMC, "  Buffer: %s -> (%p)\n", print_hex(str, buf, tx>255?255:tx), mmc);
         } else {
-            DEBUG(DBG_MMC, "  Buffer: %s <- (%p)\n", print_hex(str, buf, rx), mmc);
+            DEBUG(DBG_MMC, "  Buffer: %s <- (%p)\n", print_hex(str, buf, rx>255?255:rx), mmc);
         }
 
         if (a >= 0) {
@@ -210,6 +210,10 @@ static int _mmc_send_cmd(MMC *mmc, const uint8_t *cmd, uint8_t *buf, size_t tx,
     return 0;
 }
 
+/*
+ *
+ */
+
 static int _mmc_report_key(MMC *mmc, uint8_t agid, uint32_t addr,
                            uint8_t blocks, uint8_t format, uint8_t *buf,
                            uint16_t len)
@@ -254,6 +258,49 @@ static int _mmc_send_key(MMC *mmc, uint8_t agid, uint8_t format, uint8_t *buf,
     return _mmc_send_cmd(mmc, cmd, buf, len, 0);
 }
 
+static int _mmc_report_disc_structure(MMC *mmc, uint8_t agid, uint8_t format,
+                                      uint8_t layer, uint32_t address,
+                                      uint8_t *buf, uint16_t len)
+{
+    uint8_t cmd[16];
+    memset(cmd, 0, sizeof(cmd));
+    memset(buf, 0, len);
+
+    cmd[0] = 0xad; // operation code
+    cmd[1] = 0x01; // BluRay
+    cmd[2] = (address >> 24) & 0xff;
+    cmd[3] = (address >> 16) & 0xff;
+    cmd[4] = (address >> 8) & 0xff;
+    cmd[5] = address & 0xff;
+    cmd[6] = layer;
+    cmd[7] = format;
+    cmd[8] = (len >> 8) & 0xff;
+    cmd[9] = len & 0xff;
+    cmd[10] = (agid << 6);
+
+    return _mmc_send_cmd(mmc, cmd, buf, 0, len);
+}
+
+static int _mmc_get_configuration(MMC *mmc, uint16_t feature, uint8_t *buf, uint16_t len)
+{
+    uint8_t cmd[16];
+    memset(cmd, 0, sizeof(cmd));
+    memset(buf, 0, len);
+
+    cmd[0] = 0x46; // operation code
+    cmd[1] = 0x01; // BluRay
+    cmd[2] = (feature >> 8) & 0xff;
+    cmd[3] = feature & 0xff;
+    cmd[7] = (len >> 8) & 0xff;
+    cmd[8] = len & 0xff;
+
+    return _mmc_send_cmd(mmc, cmd, buf, 0, len);
+}
+
+/*
+ *
+ */
+
 static int _mmc_invalidate_agid(MMC *mmc, uint8_t agid)
 {
     uint8_t buf[2];
@@ -326,6 +373,64 @@ static int _mmc_read_drive_key(MMC *mmc, uint8_t agid, uint8_t *drive_key_point,
     return 0;
 }
 
+
+static int _mmc_check_aacs(MMC *mmc)
+{
+    uint8_t buf[16];
+    memset(buf, 0, sizeof(buf));
+
+    if (_mmc_get_configuration(mmc, 0x010d, buf, 16)) {
+        uint16_t feature = MKINT_BE16(buf+8);
+        if (feature == 0x010d) {
+            DEBUG(DBG_MMC, "AACS feature descriptor:\n");
+            DEBUG(DBG_MMC, "  AACS version: %d\n", buf[5+8]);
+            DEBUG(DBG_MMC, "  AACS active: %d\n", buf[2+8] & 1);
+            DEBUG(DBG_MMC, "  Binding Nonce generation support: %d\n", buf[4+8] & 1);
+            DEBUG(DBG_MMC, "  Binding Nonce block count: %d\n", buf[5+8]);
+            DEBUG(DBG_MMC, "  Bus encryption support: %d\n", buf[4+8] & 2);
+            DEBUG(DBG_MMC, "  AGID count: %d\n", buf[6+8] & 0xf);
+
+            return buf[2+8] & 1;
+        }
+        DEBUG(DBG_MMC, "incorrect feature ID %04x\n", feature);
+    }
+
+    DEBUG(DBG_MMC, "_mmc_get_configuration() failed\n");
+    return 0;
+}
+
+static uint8_t *_mmc_read_mkb(MMC *mmc, uint8_t agid, int address, int *size)
+{
+    const int layer = 0;
+    uint8_t *mkb = NULL;
+    uint8_t buf[32772];
+
+    *size = 0;
+    if (_mmc_report_disc_structure(mmc, agid, 0x83, layer, address, buf, sizeof(buf))) {
+        unsigned pack, num_packs = buf[3];
+        int32_t  len = MKINT_BE16(buf) - 2;
+        mkb = malloc(32768 * num_packs);
+
+        DEBUG(DBG_MMC, "got mkb: pack 0/%d %d bytes\n", num_packs, len);
+        memcpy(mkb, buf + 4, len);
+       *size += len;
+
+        for (pack = 1; pack < num_packs; pack++) {
+            if (_mmc_report_disc_structure(mmc, agid, 0x83, layer, pack, buf, sizeof(buf))) {
+                len = MKINT_BE16(buf) - 2;
+                DEBUG(DBG_MMC, "got mkb: pack %d/%d %d bytes\n", pack, num_packs, len);
+
+                memcpy(mkb + *size, buf + 4, len);
+                *size += len;
+            } else {
+                break;
+            }
+        }
+    }
+
+    return mkb;
+}
+
 static int _mmc_send_host_key(MMC *mmc, uint8_t agid,
                               const uint8_t *host_key_point,
                               const uint8_t *host_key_signature)
@@ -344,17 +449,9 @@ static int _mmc_read_vid(MMC *mmc, uint8_t agid, uint8_t *volume_id,
                          uint8_t *mac)
 {
     uint8_t buf[36];
-    uint8_t cmd[16];
-    memset(cmd, 0, sizeof(cmd));
-    memset(buf, 0, 36);
+    memset(buf, 0, sizeof(buf));
 
-    cmd[0] = 0xad;
-    cmd[1] = 1;    // 1 = BLURAY
-    cmd[7] = 0x80;
-    cmd[9] = 0x24;
-    cmd[10] = (agid << 6) & 0xc0;
-
-    if (_mmc_send_cmd(mmc, cmd, buf, 0, 36)) {
+    if (_mmc_report_disc_structure(mmc, agid, 0x80, 0, 0, buf, 36)) {
         memcpy(volume_id, buf + 4,  16);
         memcpy(mac,       buf + 20, 16);
         return 1;
@@ -390,7 +487,7 @@ MMC *mmc_open(const char *path)
 #ifdef HAVE_REALPATH
     char *file_path = malloc(PATH_MAX);
     if (!file_path || !realpath(path, file_path)) {
-        DEBUG(DBG_MMC, "Failed resolving path %s (%p)\n", path, mmc);
+        DEBUG(DBG_MMC | DBG_CRIT, "Failed resolving path %s (%p)\n", path, mmc);
         X_FREE(mmc);
         X_FREE(file_path);
         return NULL;
@@ -435,7 +532,7 @@ MMC *mmc_open(const char *path)
     }
 
     if (mmc->fd < 0) {
-        DEBUG(DBG_MMC, "Error opening LINUX MMC drive mounted to %s (%p)\n", file_path, mmc);
+        DEBUG(DBG_MMC | DBG_CRIT, "Error opening LINUX MMC drive mounted to %s (%p)\n", file_path, mmc);
         X_FREE(mmc);
     }
 
@@ -450,7 +547,7 @@ MMC *mmc_open(const char *path)
     UINT type = GetDriveType(drive);
 
     if (type != DRIVE_CDROM) {
-        DEBUG(DBG_MMC, "Drive %s is not CD/DVD drive !\n", drive);
+        DEBUG(DBG_MMC | DBG_CRIT, "Drive %s is not CD/DVD drive\n", drive);
         X_FREE(mmc);
         return NULL;
     }
@@ -473,11 +570,16 @@ MMC *mmc_open(const char *path)
 
 #else
 
-    DEBUG(DBG_MMC, "No MMC drive support !\n");
+    DEBUG(DBG_MMC | DBG_CRIT, "No MMC drive support\n");
     X_FREE(mmc);
 
 #endif
 
+    if (mmc && !_mmc_check_aacs(mmc)) {
+        DEBUG(DBG_MMC | DBG_CRIT, "AACS not active or supported by the drive\n");
+        X_FREE(mmc);
+    }
+
     return mmc;
 }
 
@@ -517,6 +619,7 @@ int mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cer
 {
     uint8_t agid = 0, hks[40], dn[20], dc[92], dkp[40], dks[40], mac[16];
     char str[512];
+    int error_code = MMC_ERROR;
 
     memset(hks, 0, sizeof(hks));
 
@@ -526,7 +629,7 @@ int mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cer
 
     if (!_mmc_report_agid(mmc, &agid)) {
         DEBUG(DBG_MMC | DBG_CRIT, "Didn't get AGID from drive (%p)\n", mmc);
-        return 0;
+        return MMC_ERROR;
     }
     DEBUG(DBG_MMC, "Got AGID from drive: %d (%p)\n", agid, mmc);
 
@@ -542,6 +645,7 @@ int mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cer
             DEBUG(DBG_MMC | DBG_CRIT,
                   "Host key / Certificate has been revoked by your drive ? "
                   "(%p)\n", mmc);
+            error_code = MMC_ERROR_CERT_REVOKED;
             break;
         }
 
@@ -608,12 +712,25 @@ int mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cer
 
         _mmc_invalidate_agid(mmc, agid);
 
-        return 1;
+        return MMC_SUCCESS;
     }
 
     DEBUG(DBG_MMC | DBG_CRIT, "Unable to read VID from drive! (%p)\n", mmc);
 
     _mmc_invalidate_agid(mmc, agid);
 
-    return 0;
+    return error_code;
+}
+
+uint8_t *mmc_read_mkb(MMC *mmc, int address, int *size)
+{
+    uint8_t agid = 0;
+
+    _mmc_report_agid(mmc, &agid);
+
+    uint8_t *mkb = _mmc_read_mkb(mmc, agid, address, size);
+
+    _mmc_invalidate_agid(mmc, agid);
+
+    return mkb;
 }
